@@ -1,44 +1,66 @@
-import schedule
-from typing import Dict, Any, Generator, Iterable
+from typing import Dict, Any, Generator, Iterable, NoReturn, Optional
 from copy import deepcopy
 from fastapi import HTTPException
-from sovereign import config
 from sovereign.config_loader import Loadable
 from sovereign.schemas import DiscoveryRequest, XdsTemplate
-from sovereign.sources import extract_node_key, get_instances_for_node
-from sovereign.utils import crypto
-from sovereign.utils.crypto import disabled_suite
-
-REFRESH_RATE = config.template_context.refresh_rate
-REFRESH_ENABLED = config.template_context.refresh
-CONFIGURED_CONTEXT = config.template_context.context
+from sovereign.sources import SourcePoller
+from sovereign.utils.crypto import CipherSuite
+from sovereign.utils.timer import poll_forever, poll_forever_cron
 
 
 class TemplateContext:
-    def __init__(self) -> None:
+    def __init__(
+        self,
+        refresh_rate: Optional[int],
+        refresh_cron: Optional[str],
+        configured_context: Dict[str, Loadable],
+        poller: SourcePoller,
+        encryption_suite: CipherSuite,
+        disabled_suite: CipherSuite,
+        logger: Any,
+        stats: Any,
+    ) -> None:
+        self.poller = poller
+        self.refresh_rate = refresh_rate
+        self.refresh_cron = refresh_cron
+        self.configured_context = configured_context
+        self.crypto = encryption_suite
+        self.disabled_suite = disabled_suite
+        # initial load
         self.context = self.load_context_variables()
-        self.refresh_context()  # One-time setup
-        if REFRESH_ENABLED:
-            # Continuous re-loading of context variables
-            schedule.every(REFRESH_RATE).seconds.do(self.refresh_context)
+        self.logger = logger
+        self.stats = stats
 
-    def refresh_context(self) -> None:
-        self.context = self.load_context_variables()
+    async def start_refresh_context(self) -> NoReturn:
+        if self.refresh_cron is not None:
+            await poll_forever_cron(self.refresh_cron, self.refresh_context)
+        elif self.refresh_rate is not None:
+            await poll_forever(self.refresh_rate, self.refresh_context)
 
-    @staticmethod
-    def load_context_variables() -> Dict[str, Any]:
+        raise RuntimeError("Failed to start refresh_context, this should never happen")
+
+    async def refresh_context(self) -> None:
+        try:
+            self.context = self.load_context_variables()
+            self.stats.increment("context.refresh.success")
+        # pylint: disable=broad-except
+        except Exception as e:
+            self.logger(event=e)
+            self.stats.increment("context.refresh.error")
+
+    def load_context_variables(self) -> Dict[str, Any]:
         ret = dict()
-        for k, v in CONFIGURED_CONTEXT.items():
+        for k, v in self.configured_context.items():
             if isinstance(v, Loadable):
                 ret[k] = v.load()
             elif isinstance(v, str):
                 ret[k] = Loadable.from_legacy_fmt(v).load()
         if "crypto" not in ret:
-            ret["crypto"] = crypto
+            ret["crypto"] = self.crypto
         return ret
 
     def build_new_context_from_instances(self, node_value: str) -> Dict[str, Any]:
-        matches = get_instances_for_node(node_value=node_value)
+        matches = self.poller.match_node(node_value=node_value)
         ret = dict()
         for key, value in self.context.items():
             try:
@@ -66,26 +88,16 @@ class TemplateContext:
         ret.update(to_add)
         return ret
 
-    def safe(self, request: DiscoveryRequest) -> Dict[str, Any]:
-        ret = self.build_new_context_from_instances(
-            node_value=extract_node_key(request.node),
-        )
-        # If the discovery request came from a mock, it will
-        # typically contain this metadata key.
-        # This means we should prevent any decryptable data
-        # from ending up in the response.
-        if request.hide_private_keys:
-            ret["crypto"] = disabled_suite
-        return ret
-
     def get_context(
         self, request: DiscoveryRequest, template: XdsTemplate
     ) -> Dict[str, Any]:
-        ret = self.safe(request)
+        ret = self.build_new_context_from_instances(
+            node_value=self.poller.extract_node_key(request.node),
+        )
+        if request.hide_private_keys:
+            ret["crypto"] = self.disabled_suite
         if not template.is_python_source:
-            keys_to_remove = self.unused_variables(
-                list(ret), template.jinja_variables()
-            )
+            keys_to_remove = self.unused_variables(list(ret), template.jinja_variables)
             for key in keys_to_remove:
                 ret.pop(key, None)
         return ret
@@ -100,6 +112,3 @@ class TemplateContext:
 
     def get(self, *args: Any, **kwargs: Any) -> Any:
         return self.context.get(*args, **kwargs)
-
-
-template_context = TemplateContext()
