@@ -1,16 +1,24 @@
-import zlib
 import warnings
 import multiprocessing
 from collections import defaultdict
-from datetime import datetime, timedelta
 from enum import Enum
-from pydantic import BaseModel, Field, BaseSettings, SecretStr, validator
-from typing import List, Any, Dict, Union, Optional, Tuple, Sized, Type
+from pydantic import (
+    BaseModel,
+    Field,
+    BaseSettings,
+    SecretStr,
+    validator,
+    root_validator,
+)
+from typing import List, Any, Dict, Union, Optional, Tuple, Type
+from types import ModuleType
 from jinja2 import meta, Template
 from fastapi.responses import JSONResponse
 from sovereign.config_loader import jinja_env, Serialization, Protocol, Loadable
 from sovereign.utils.version_info import compute_hash
+from croniter import croniter, CroniterBadCronError
 
+missing_arguments = {"missing", "positional", "arguments:"}
 
 JsonResponseClass: Type[JSONResponse] = JSONResponse
 # pylint: disable=unused-import
@@ -44,27 +52,6 @@ class ConfiguredSource(BaseModel):
     scope: str = "default"  # backward compatibility
 
 
-class SourceMetadata(BaseModel):
-    updated: datetime = datetime.fromtimestamp(0)
-    count: int = 0
-
-    def update_date(self) -> None:
-        self.updated = datetime.now()
-
-    def update_count(self, iterable: Sized) -> None:
-        self.count = len(iterable)
-
-    @property
-    def is_stale(self) -> bool:
-        return self.updated < (datetime.now() - timedelta(minutes=2))
-
-    def __str__(self) -> str:
-        return (
-            f"Sources were last updated at {datetime.isoformat(self.updated)}. "
-            f"There are {self.count} instances."
-        )
-
-
 class StatsdConfig(BaseModel):
     host: str = "127.0.0.1"
     port: int = 8125
@@ -94,14 +81,20 @@ class XdsTemplate:
             self.loadable = path
         self.is_python_source = self.loadable.protocol == Protocol.python
         self.source = self.load_source()
-        self.checksum = zlib.adler32(self.source.encode())
+        template_ast = jinja_env.parse(self.source)
+        self.jinja_variables = meta.find_undeclared_variables(template_ast)  # type: ignore
 
-    async def __call__(self, *args: Any, **kwargs: Any) -> Union[Dict[str, Any], str]:
-        if self.is_python_source:
-            code = self.loadable.load()
+    def __call__(
+        self, *args: Any, **kwargs: Any
+    ) -> Optional[Union[Dict[str, Any], str]]:
+        if not hasattr(self, "code"):
+            self.code: Union[Template, ModuleType] = self.loadable.load()
+        if isinstance(self.code, ModuleType):
             try:
-                return {"resources": list(code.call(*args, **kwargs))}
+                return {"resources": list(self.code.call(*args, **kwargs))}  # type: ignore
             except TypeError as e:
+                if not set(str(e).split()).issuperset(missing_arguments):
+                    raise e
                 message_start = str(e).find(":")
                 missing_args = str(e)[message_start + 2 :]
                 supplied_args = list(kwargs.keys())
@@ -111,12 +104,7 @@ class XdsTemplate:
                     f"Supplied keyword args: {supplied_args}"
                 )
         else:
-            content: Template = self.loadable.load()
-            return await content.render_async(*args, **kwargs)
-
-    def jinja_variables(self) -> List[str]:
-        template_ast = jinja_env.parse(self.source)
-        return meta.find_undeclared_variables(template_ast)  # type: ignore
+            return self.code.render(*args, **kwargs)
 
     def load_source(self) -> str:
         if self.loadable.serialization in (Serialization.jinja, Serialization.jinja2):
@@ -142,24 +130,22 @@ class XdsTemplate:
             self.loadable.protocol = old_protocol
             self.loadable.serialization = old_serialization
             return str(ret)
-        else:
-            # The only other supported serializers are string, yaml, and json
-            # So it should be safe to create this checksum off
-            return str(self.source)
+        ret = self.loadable.load()
+        return str(ret)
+
+    def __repr__(self) -> str:
+        return f"XdsTemplate({self.loadable=}, {self.is_python_source=}, {self.source=}, {self.jinja_variables=})"
 
 
 class ProcessedTemplate:
     def __init__(
         self,
         resources: List[Dict[str, Any]],
-        type_url: str,
         version_info: Optional[str],
     ) -> None:
-        for resource in resources:
-            if "@type" not in resource:
-                resource["@type"] = type_url
         self.resources = resources
         self.version_info = version_info
+        self._rendered: Optional[bytes] = None
 
     @property
     def version(self) -> str:
@@ -167,25 +153,20 @@ class ProcessedTemplate:
 
     @property
     def rendered(self) -> bytes:
-        return JsonResponseClass().render(
-            content={
-                "version_info": self.version,
-                "resources": self.resources,
-            }
-        )
+        if self._rendered is None:
+            result = JsonResponseClass().render(
+                content={
+                    "version_info": self.version,
+                    "resources": self.resources,
+                }
+            )
+            self._rendered = result
+        return self._rendered
 
     def deserialize_resources(self) -> List[Dict[str, Any]]:
         return self.resources
 
 
-class ProcessedTemplates:
-    def __init__(self, types: Optional[Dict[str, ProcessedTemplate]] = None) -> None:
-        if types is None:
-            self.types = dict()
-        else:
-            self.types = types
-
-
 class Locality(BaseModel):
     region: str = Field(None)
     zone: str = Field(None)
@@ -270,7 +251,7 @@ class DiscoveryRequest(BaseModel):
         Resources(), title="List of requested resource names"
     )
     hide_private_keys: bool = False
-    type_url: str = Field(
+    type_url: Optional[str] = Field(
         None, title="The corresponding type_url for the requested resource"
     )
     desired_controlplane: str = Field(
@@ -369,7 +350,8 @@ class SovereignConfig(BaseSettings):
     sources_refresh_rate: int = 30
     cache_strategy: str = "context"
     refresh_context: bool = False
-    context_refresh_rate: int = 3600
+    context_refresh_rate: Optional[int]
+    context_refresh_cron: Optional[str]
     dns_hard_fail: bool = False
     enable_application_logs: bool = False
     enable_access_logs: bool = True
@@ -391,6 +373,7 @@ class SovereignConfig(BaseSettings):
             "cache_strategy": {"env": "SOVEREIGN_CACHE_STRATEGY"},
             "refresh_context": {"env": "SOVEREIGN_REFRESH_CONTEXT"},
             "context_refresh_rate": {"env": "SOVEREIGN_CONTEXT_REFRESH_RATE"},
+            "context_refresh_cron": {"env": "SOVEREIGN_CONTEXT_REFRESH_CRON"},
             "dns_hard_fail": {"env": "SOVEREIGN_DNS_HARD_FAIL"},
             "enable_application_logs": {"env": "SOVEREIGN_ENABLE_APPLICATION_LOGS"},
             "enable_access_logs": {"env": "SOVEREIGN_ENABLE_ACCESS_LOGS"},
@@ -492,7 +475,8 @@ class LoggingConfiguration(BaseSettings):
 class ContextConfiguration(BaseSettings):
     context: Dict[str, Loadable] = {}
     refresh: bool = False
-    refresh_rate: int = 3600
+    refresh_rate: Optional[int] = None
+    refresh_cron: Optional[str] = None
 
     @staticmethod
     def context_from_legacy(context: Dict[str, str]) -> Dict[str, Loadable]:
@@ -501,10 +485,41 @@ class ContextConfiguration(BaseSettings):
             ret[key] = Loadable.from_legacy_fmt(value)
         return ret
 
+    @root_validator(pre=False)
+    def validate_single_use_refresh_method(
+        cls, values: Dict[str, Any]
+    ) -> Dict[str, Any]:
+        refresh_rate = values.get("refresh_rate")
+        refresh_cron = values.get("refresh_cron")
+
+        if (refresh_rate is not None) and (refresh_cron is not None):
+            raise RuntimeError(
+                f"Only one of SOVEREIGN_CONTEXT_REFRESH_RATE or SOVEREIGN_CONTEXT_REFRESH_CRON can be defined. Got {refresh_rate=} and {refresh_cron=}"
+            )
+        return values
+
+    @root_validator
+    def set_default_refresh_rate(cls, values: Dict[str, Any]) -> Dict[str, Any]:
+        refresh_rate = values.get("refresh_rate")
+        refresh_cron = values.get("refresh_cron")
+
+        if (refresh_rate is None) and (refresh_cron is None):
+            values["refresh_rate"] = 3600
+        return values
+
+    @validator("refresh_cron")
+    def validate_refresh_cron(cls, v: Optional[str]) -> Optional[str]:
+        if v is None:
+            return v
+        if not croniter.is_valid(v):
+            raise CroniterBadCronError(f"'{v}' is not a valid cron expression")
+        return v
+
     class Config:
         fields = {
             "refresh": {"env": "SOVEREIGN_REFRESH_CONTEXT"},
             "refresh_rate": {"env": "SOVEREIGN_CONTEXT_REFRESH_RATE"},
+            "refresh_cron": {"env": "SOVEREIGN_CONTEXT_REFRESH_CRON"},
         }
 
 
@@ -656,7 +671,7 @@ class SovereignConfigv2(BaseSettings):
             templates=new_templates,
             source_config=SourcesConfiguration(
                 refresh_rate=other.sources_refresh_rate,
-                cache_strategy=other.cache_strategy,
+                cache_strategy=CacheStrategy(other.cache_strategy),
             ),
             modifiers=other.modifiers,
             global_modifiers=other.global_modifiers,
@@ -666,6 +681,7 @@ class SovereignConfigv2(BaseSettings):
                 ),
                 refresh=other.refresh_context,
                 refresh_rate=other.context_refresh_rate,
+                refresh_cron=other.context_refresh_cron,
             ),
             matching=NodeMatching(
                 enabled=other.node_matching,
